'Test case operations'

from pathlib import PurePosixPath
import re
import shlex

from docker.errors import NotFound

from .dockerpy import exec_run, get_bin, put_bin
from .status import Status


def __init__(container, processor, i, ioput, config):
    'Copy binary files to `i` and judge'
    container.exec_run(f'cp -r 0 {i}', workdir=str(processor.workdir))
    exec_run(container, processor.before_judge, f'{processor.workdir}/{i}')
    res = judge(container, processor, i, ioput, config)
    exec_run(container, processor.after_judge, f'{processor.workdir}/{i}')
    return res


def _get_io_file_path(ioro, processor, i, config):
    'Get the absolute path of input or output file'
    return PurePosixPath(
        f"{processor.workdir}/{i}/{config['iofilename'][ioro]}"
        if ioro in config.get('iofilename', {})
        else f'{processor.workdir}/{i}.{ioro}'
    )


def judge(container, processor, i, ioput, config):
    'Judge one of the test cases'
    put_bin(container, _get_io_file_path('in', processor, i, config), ioput[0])
    res = container.exec_run(
        'bash -c ' + shlex.quote(
            "TIMEFORMAT=$'\\n%3lR' && time timeout -sKILL "
            + str(config.get('limit', {}).get('time', 1))
            + ' sh -c ' + shlex.quote(processor.judge)
            + ' > ' + f'{processor.workdir}/{i}.out'
            + (' < ' + f'{processor.workdir}/{i}.in'
               if 'in' not in config.get('iofilename', {}) else '')
        ),
        workdir=f'{processor.workdir}/{i}', demux=True
    )
    duration = re.search('\n([0-9]+)m([0-9]+\\.[0-9]{3})s\n$',
                         res.output[1].decode())
    stderr = res.output[1][:duration.span()[0]]
    duration = int(duration.group(1)) * 60 + float(duration.group(2))
    if res.exit_code == 137:
        return Status.TLE, (None, stderr), duration
    if res.exit_code:
        return Status.RE, (None, stderr), duration
    try:
        output = get_bin(container,
                         _get_io_file_path('out', processor, i, config))
    except NotFound:
        return Status.ONF, (None, stderr), duration
    return (Status.AC if output.rstrip() == ioput[1].rstrip() else Status.WA,
            (output, stderr), duration)



'Processors'
# pylint: disable = missing-function-docstring, too-few-public-methods

from enum import Enum
from pathlib import PurePosixPath


class Processor():
    'Defines the operations of a multi-version programming language processor'

    @staticmethod
    def _get_image_with_tag(image, tag):
        return image + (f':{tag}' if tag else '')

    image = None
    workdir = PurePosixPath('/dockerjudge')
    source = None
    before_compile = None
    compile = None
    after_compile = None
    before_judge = None
    judge = None
    after_judge = None


class _Language(Enum):
    'Get programming language from enum'

    @classmethod
    def __get_language(cls, language, default=None):
        if isinstance(language, cls):
            return language
        try:
            return cls[language]
        except KeyError:
            try:
                return cls(language)
            except ValueError:
                return cls.__get_language(default)


class Bash(Processor):
    "Bash is the GNU Project's Bourne Again SHell"

    def __init__(self, version=None):
        self.image = self._get_image_with_tag('bash', version)
        self.source = 'bash.sh'
        self.compile = ['bash', '-n', self.source]
        self.judge = f'sh {self.source}'


class Clang(Processor):
    'Clang C Language Family Frontend for LLVM'

    class Language(_Language):
        'Programming language, C (c) or C++ (cpp)'
        c = 'C'
        cpp = 'C++'

        @classmethod
        def _get_language(cls, language):
            return super().__get_language(language, cls.cpp)

    def __init__(self, language=None, version=None,
                 filenames=None, options=None):
        lang = self.Language._get_language(language)
        fns = filenames or {}
        args = options or []

        self.image = 'clangbuiltlinux/ubuntu' + f':llvm{version}-latest'
        self.source = fns.get('src', f'a.{lang.name}')
        self.compile = ([{self.Language.c: 'clang',
                          self.Language.cpp: 'clang++'}[lang] + f'-{version}',
                         self.source]
                        + (['-o', fns['bin']]
                           if fns.get('bin') else []) + args)
        self.after_compile = ['rm', self.source]
        self.judge = f"./{fns.get('bin', 'a.out')}"


class GCC(Processor):
    'GNU project C, C++ and Go compiler'

    class Language(_Language):
        'Programming language, C, C++ or Go'
        c = 'C'
        cpp = 'C++'
        go = 'Go'

        @classmethod
        def _get_language(cls, language):
            return super().__get_language(language, cls.cpp)

    def __init__(self, language=None, version=None,
                 filenames=None, options=None):
        lang = self.Language._get_language(language)
        fns = filenames or {}
        args = options or []

        self.image = self._get_image_with_tag('gcc', version)
        self.source = fns.get('src', f'a.{lang.name}')
        self.compile = ([{self.Language.c: 'gcc',
                          self.Language.cpp: 'g++',
                          self.Language.go: 'gccgo'}[lang],
                         self.source]
                        + (['-o', fns['bin']]
                           if fns.get('bin') else []) + args)
        self.after_compile = ['rm', self.source]
        self.judge = f"./{fns.get('bin', 'a.out')}"


class Go(Processor):
    'The Go Programming Language'

    def __init__(self, version=None, filenames=None, options=None):
        fns = filenames or {}
        args = options or []

        self.image = self._get_image_with_tag('golang', version)
        self.source = fns.get('src', 'main.go')
        self.compile = (['go', 'build']
                        + (['-o', fns['bin']]
                           if fns.get('bin') else []) + args
                        + [self.source])
        self.after_compile = ['rm', self.source]
        self.judge = f"./{fns.get('bin', 'main')}"


class Node(Processor):
    'Node.js®'

    def __init__(self, version=None):
        self.image = self._get_image_with_tag('node', version)
        self.source = 'index.js'
        self.compile = ['node', '-c', self.source]
        self.judge = f'node {self.source}'


class OpenJDK(Processor):
    'Open Java Development Kit'

    def __init__(self, version=None):
        self.image = self._get_image_with_tag('openjdk', version)
        self.source = 'Main.java'
        self.compile = ['javac', self.source]
        self.after_compile = ['rm', self.source]
        self.judge = 'java Main'


class Python(Processor):
    'CPython'

    def __init__(self, version=None):
        self.image = self._get_image_with_tag('python', version)
        self.source = '__init__.py'
        self.compile = ['python', '-m', 'compileall', '.']
        self.judge = f'python {self.source}'



'The collection of judge statuses'

from enum import Enum


class Status(Enum):
    'Enumeration of judge statuses'
    AC = 'Accepted'
    WA = 'Wrong Answer'
    ONF = 'Output Not Found'
    RE = 'Runtime Error'
    TLE = 'Time Limit Exceeded'
    UE = 'Unknown Error'
    CE = 'Compilation Error'
